@@ -5,7 +5,7 @@
  * Documented by mathew <meta@pobox.com>
  *
  * $RoughId: syslog.c,v 1.21 2002/02/25 12:21:17 knu Exp $
- * $Id: syslog.c 47792 2014-10-04 23:09:32Z nobu $
+ * $Id$
  */
 
 #include "ruby/ruby.h"
@@ -150,6 +150,7 @@ static VALUE mSyslog_close(VALUE self)
 static VALUE mSyslog_open(int argc, VALUE *argv, VALUE self)
 {
     VALUE ident, opt, fac;
+    const char *ident_ptr;
 
     if (syslog_opened) {
         rb_raise(rb_eRuntimeError, "syslog already open");
@@ -160,8 +161,8 @@ static VALUE mSyslog_open(int argc, VALUE *argv, VALUE self)
     if (NIL_P(ident)) {
         ident = rb_gv_get("$0");
     }
-    SafeStringValue(ident);
-    syslog_ident = strdup(RSTRING_PTR(ident));
+    ident_ptr = StringValueCStr(ident);
+    syslog_ident = strdup(ident_ptr);
 
     if (NIL_P(opt)) {
 	syslog_options = LOG_PID | LOG_CONS;
@@ -295,18 +296,12 @@ static VALUE mSyslog_set_mask(VALUE self, VALUE mask)
  *   Syslog.log(Syslog::LOG_ALERT, "Out of memory")
  *   Syslog.alert("Out of memory")
  *
- * Format strings are as for printf/sprintf, except that in addition %m is
- * replaced with the error message string that would be returned by
- * strerror(errno).
- *
  */
 static VALUE mSyslog_log(int argc, VALUE *argv, VALUE self)
 {
     VALUE pri;
 
-    if (argc < 2) {
-        rb_raise(rb_eArgError, "wrong number of arguments (%d for 2+)", argc);
-    }
+    rb_check_arity(argc, 2, UNLIMITED_ARGUMENTS);
 
     argc--;
     pri = *argv++;
@@ -420,6 +415,7 @@ static VALUE mSyslogMacros_included(VALUE mod, VALUE target)
  */
 void Init_syslog(void)
 {
+#undef rb_intern
     mSyslog = rb_define_module("Syslog");
 
     mSyslogConstants    = rb_define_module_under(mSyslog, "Constants");
@@ -506,7 +502,7 @@ void Init_syslog(void)
     rb_define_syslog_facility(LOG_NEWS);
 #endif
 #ifdef LOG_NTP
-   rb_define_syslog_facility(LOG_NTP);
+    rb_define_syslog_facility(LOG_NTP);
 #endif
 #ifdef LOG_SECURITY
     rb_define_syslog_facility(LOG_SECURITY);
