@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 require_relative 'helper'
 
 module Psych
@@ -14,6 +15,32 @@ module Psych
     def setup
       super
       @wups = Wups.new
+
+      @orig_verbose, $VERBOSE = $VERBOSE, nil
+    end
+
+    def teardown
+      $VERBOSE = @orig_verbose
+    end
+
+    def make_ex msg = 'oh no!'
+      begin
+        raise msg
+      rescue ::Exception => e
+        e
+      end
+    end
+
+    def test_backtrace
+      err     = make_ex
+      new_err = Psych.load(Psych.dump(err))
+      assert_equal err.backtrace, new_err.backtrace
+    end
+
+    def test_naming_exception
+      err     = String.xxx rescue $!
+      new_err = Psych.load(Psych.dump(err))
+      assert_equal err.message, new_err.message
     end
 
     def test_load_takes_file
@@ -23,9 +50,15 @@ module Psych
       assert_nil ex.file
 
       ex = assert_raises(Psych::SyntaxError) do
-        Psych.load '--- `', 'meow'
+        Psych.load '--- `', filename: 'meow'
       end
       assert_equal 'meow', ex.file
+
+      # deprecated interface
+      ex = assert_raises(Psych::SyntaxError) do
+        Psych.load '--- `', 'deprecated'
+      end
+      assert_equal 'deprecated', ex.file
     end
 
     def test_psych_parse_stream_takes_file
@@ -36,7 +69,7 @@ module Psych
       assert_match '(<unknown>)', ex.message
 
       ex = assert_raises(Psych::SyntaxError) do
-        Psych.parse_stream '--- `', 'omg!'
+        Psych.parse_stream '--- `', filename: 'omg!'
       end
       assert_equal 'omg!', ex.file
       assert_match 'omg!', ex.message
@@ -50,9 +83,15 @@ module Psych
       assert_match '(<unknown>)', ex.message
 
       ex = assert_raises(Psych::SyntaxError) do
-        Psych.load_stream '--- `', 'omg!'
+        Psych.load_stream '--- `', filename: 'omg!'
       end
       assert_equal 'omg!', ex.file
+
+      # deprecated interface
+      ex = assert_raises(Psych::SyntaxError) do
+        Psych.load_stream '--- `', 'deprecated'
+      end
+      assert_equal 'deprecated', ex.file
     end
 
     def test_parse_file_exception
@@ -87,9 +126,15 @@ module Psych
       assert_nil ex.file
 
       ex = assert_raises(Psych::SyntaxError) do
-        Psych.parse '--- `', 'omg!'
+        Psych.parse '--- `', filename: 'omg!'
       end
       assert_match 'omg!', ex.message
+
+      # deprecated interface
+      ex = assert_raises(Psych::SyntaxError) do
+        Psych.parse '--- `', 'deprecated'
+      end
+      assert_match 'deprecated', ex.message
     end
 
     def test_attributes
@@ -114,19 +159,6 @@ module Psych
       assert_equal 2, w.bar
     end
 
-    def test_to_yaml_properties
-      class << @wups
-        def to_yaml_properties
-          [:@foo]
-        end
-      end
-
-      w = Psych.load(Psych.dump(@wups))
-      assert_equal @wups, w
-      assert_equal 1, w.foo
-      assert_nil w.bar
-    end
-
     def test_psych_syntax_error
       Tempfile.create(['parsefile', 'yml']) do |t|
         t.binmode
